import { cartesianProduct } from 'cartesian-product-multiple-arrays';
import {
    FundingSource, BarChartDataset,
    BasicDataset, NrenStaff, TrafficVolume, NrenAndYearDatapoint,
    TrafficRatio,
    URLDatapoint
} from "compendium/Schema";

// create a color from a string, credits https://stackoverflow.com/a/16348977
const stringToColour = function (str) {
    let hash = 0;
    for (let i = 0; i < str.length; i++) {
        hash = str.charCodeAt(i) + ((hash << 5) - hash);
    }
    let colour = '#';
    for (let i = 0; i < 3; i++) {
        const value = (hash >> (i * 8)) & 0xFF;
        const valueAsString = '00' + value.toString(16);
        colour += valueAsString.substring(valueAsString.length - 2);
    }
    return colour;
}

export function addTooltip<T extends NrenAndYearDatapoint>(dataLookup: Map<string, Map<string, Map<number, T>>>, processTooltip = (_column: string, _datapoint: T): string | undefined => undefined) {
    const withTooltip = new Map<string, Map<string, Map<number, { [key: string]: string | number }>>>();

    for (const [nren, nrenMap] of dataLookup) {
        const newNrenMap = new Map<string, Map<number, { [key: string]: string | number }>>();
        for (const [column, yearMap] of nrenMap) {
            const newYearMap = new Map<number, { [key: string]: string | number }>();
            for (const [year, datapoint] of yearMap) {

                const tooltip = processTooltip(column, datapoint);

                if (tooltip) {
                    newYearMap.set(year, { tooltip: tooltip });
                    continue;
                }

                newYearMap.set(year, {});
            }
            newNrenMap.set(column, newYearMap);
        }
        withTooltip.set(nren, newNrenMap);
    }
    return withTooltip;
}

export function getLatestData<T extends NrenAndYearDatapoint>(data: T[]) {
    // Get the latest year for each NREN, and return the data for that year
    // Mainly used for URL data, which doesn't make sense to compare across years

    const latestData = new Map<string, T>();
    data.forEach(datapoint => {
        const existingData = latestData.get(datapoint.nren);
        if (!existingData || existingData.year < datapoint.year) {
            latestData.set(datapoint.nren, datapoint);
        }
    });
    return Array.from(latestData.values());
}

export function addHttpIfMissing(url: string) {
    // check if it has a protocol, if not add http (href becomes relative to the current page, so we need to add the protocol to make it work properly)
    if (!url.match(/^[a-zA-Z]+:\/\//)) {
        return 'https://' + url;
    }
    return url;
}

export const extractUrls = (datapoint: URLDatapoint) => {
    const result: { [key: string]: string } = {}

    if (!datapoint.urls && !datapoint.url) return result;

    if (datapoint.urls) {
        datapoint.urls.forEach(url => {
            result[url] = url;
        });
    }

    if (datapoint.url) {
        result[datapoint.url] = datapoint.url;
    }

    return result;
}

export function createDataLookupList<T extends NrenAndYearDatapoint>(entries: T[]) {
    const map = new Map<string, Map<number, T[]>>();
    entries.forEach(entry => {
        let nrenEntry = map.get(entry.nren);
        if (!nrenEntry) {
            nrenEntry = new Map<number, T[]>();
        }
        let list = nrenEntry.get(entry.year);
        if (!list) {
            list = [];
        }
        list.push(entry);
        nrenEntry.set(entry.year, list);
        map.set(entry.nren, nrenEntry);
    });
    return map;
}

export function createDataLookup<T extends NrenAndYearDatapoint>(entries: T[]) {
    const map = new Map<string, Map<number, T>>();
    entries.forEach(datapoint => {
        let dataForNren = map.get(datapoint.nren);
        if (!dataForNren) {
            dataForNren = new Map<number, T>();
        }
        dataForNren.set(datapoint.year, datapoint);
        map.set(datapoint.nren, dataForNren);
    });
    return map;
}

export function getTableData<T>(dataLookup: Map<string, Map<number, T>>, mutate: (obj: { [key: string]: string }, data: T) => void) {

    const dataByYear = new Map<string, Map<number, { [key: string]: string }>>();
    dataLookup.forEach((yearMap, nren) => {
        const dataByYearForNren = new Map<number, { [key: string]: string }>();

        const sortedYears = Array.from(yearMap.keys()).sort((a, b) => b - a);

        sortedYears.forEach(year => {
            const data = yearMap.get(year)!;
            const values = dataByYearForNren.get(year) || {};
            mutate(values, data);

            if (Object.keys(values).length > 0) {
                dataByYearForNren.set(year, values);
            }
        });
        dataByYear.set(nren, dataByYearForNren);
    });
    return dataByYear;
}

export function createNrenTableLookup<T extends NrenAndYearDatapoint>(data: T[], columnProperty: keyof T, nullsMatter: boolean = false) {

    // creates a lookup from NREN -> category/column -> year -> data for that year
    const dataLookup = new Map<string, Map<string, Map<number, T>>>();

    data.forEach(datapoint => {

        const processColumn = (column: string) => {
            let dataForNren = dataLookup.get(datapoint.nren);
            if (!dataForNren) {
                dataForNren = new Map<string, Map<number, T>>();
            }

            let dataForColumn = dataForNren.get(column);
            if (!dataForColumn) {
                dataForColumn = new Map<number, T>();
            }
            dataForColumn.set(datapoint.year, datapoint);
            dataForNren.set(column, dataForColumn);
            dataLookup.set(datapoint.nren, dataForNren);
        }

        let value = datapoint[columnProperty] as string | boolean | string[];
        if (typeof value === "boolean") {
            value = value ? "True" : "False";
        }
        if (nullsMatter && value == null) {
            // stringify the null to avoid issues because it's falsy
            value = `${value}`;
        }

        if (Array.isArray(value)) {
            value.forEach(processColumn);
        }
        else {
            processColumn(value);
        }
    });
    return dataLookup;
}

export function createCategoryMatrixLookup<T extends NrenAndYearDatapoint>(
    data: T[],
    valueFields: string[],
    categoryField?: keyof T,
    valueIsField: boolean = true,
    preprocess?: (data: T) => T) {
    /*
        Creates a map from a number of categories (determined by the categoryFields) to NREN -> Year -> Data for that year
        This datastructure is primarily used for the ScrollableMatrix.
    */

    // Row identifier -> NREN -> Year -> Data for that year
    const dataLookup = new Map<string, Map<number, Map<string, { [key: string]: { [key: string]: string | number } }>>>();

    const processMatrixLookup = (data: T[], categoryField: keyof T | undefined, field: string) => {
        data.forEach(datapoint => {
            if (preprocess) {
                preprocess(datapoint);
            }
            let rowIdentifier = categoryField ? datapoint[categoryField] : field;
            if (typeof rowIdentifier === "boolean") {
                rowIdentifier = rowIdentifier ? "True" : "False";
            }
            const nren = datapoint.nren;
            const year = datapoint.year;

            const nrenData = dataLookup.get(nren) || new Map(); // NREN -> Year -> Data for that year
            const yearData = nrenData.get(year) || new Map(); // Year -> Data for that year
            const values = yearData.get(rowIdentifier as string) || {}; // Data for that year

            const value = datapoint[field];
            if (value === null || value === undefined) {
                return;
            }

            const keyToUse = valueIsField ? value : field;
            const _values = values[keyToUse] || {};
            _values[`${value}`] = value;
            values[keyToUse] = _values;

            yearData.set(rowIdentifier as string, values);
            nrenData.set(year, yearData);
            dataLookup.set(nren, nrenData);
        });
    }

    if (categoryField) {
        // the categories are the values from a specific field
        for (const valueField of valueFields) {
            processMatrixLookup(data, categoryField, valueField);
        }

    } else {
        // use the valueFields as the categories themselves
        for (const valueField of valueFields) {
            processMatrixLookup(data, undefined, valueField);
        }
    }
    return dataLookup;
}

export const createTrafficVolumeDataset = (data: TrafficVolume[]) => {
    const dataLookup = createDataLookup(data);

    const labelsYear = [...new Set(data.map((item: TrafficVolume) => item.year))].sort();
    const labelsNREN = [...new Set(data.map((item: TrafficVolume) => item.nren))].sort();

    const sets = labelsYear.map(year => {
        return {
            backgroundColor: 'rgba(40, 40, 250, 0.8)',
            data: labelsNREN.map((nren) => dataLookup.get(nren)!.get(year)?.from_customers ?? null),  // we ignore the rest of the data for now..
            label: year.toString(),
            borderSkipped: true,
            barPercentage: 0.8,
            borderWidth: 0.5,
            categoryPercentage: 0.8,
            hidden: false,
            datalabels: {
                display: true,
                color: 'grey',
                formatter: function (_value, context) {
                    return context.dataset.label;
                },
                anchor: 'start',
                align: 'end',
                offset: function (context) {
                    return context.chart.chartArea.width;
                }
            }
        }
    });

    const dataResponse: BasicDataset = {
        datasets: sets,
        labels: labelsNREN.map(l => l.toString())
    }
    return dataResponse;
}

export const createFundingSourceDataset = (data: FundingSource[]) => {
    function getColorMap() {
        const rgbToHex = (r: number, g: number, b: number) => '#' + [r, g, b].map(x => {
            const hex = x.toString(16)
            return hex.length === 1 ? '0' + hex : hex
        }).join('')

        const colorMap = new Map<string, string>();
        colorMap.set("client_institutions", rgbToHex(157, 40, 114))
        colorMap.set("commercial", rgbToHex(241, 224, 79))
        colorMap.set("european_funding", rgbToHex(219, 42, 76))
        colorMap.set("gov_public_bodies", rgbToHex(237, 141, 24))
        colorMap.set("other", rgbToHex(137, 166, 121))
        return colorMap
    }

    const dataLookup = createDataLookup(data);
    const colorMap = getColorMap();

    const labelsYear = [...new Set(data.map((item: FundingSource) => item.year))].sort();
    const labelsNREN = [...new Set(data.map((item: FundingSource) => item.nren))].sort();

    const fundingSources = {
        "client_institutions": "Client Institutions",
        "commercial": "Commercial",
        "european_funding": "European Funding",
        "gov_public_bodies": "Government/Public Bodies",
        "other": "Other",
    }

    const keys = Object.keys(fundingSources);

    const fundingSourcesPerYear: { [key: string]: { [key: string]: number } } = cartesianProduct(Object.keys(fundingSources), labelsYear)
        .reduce((obj, [fundingSource, year]) => {
            const key = `${fundingSource},${year}`
            obj[key] = {};
            return obj
        }, {});

    dataLookup.forEach((yearMap, nren) => {
        yearMap.forEach((fundingSource, year) => {
            const valuesPerSource = keys.map(source => {
                const key = source;
                return fundingSource[key] || 0;
            });
            const sum = valuesPerSource.reduce((a, b) => a + b, 0);
            if (sum === 0) {
                return;
            }
            for (const fundingSource of keys) {
                const key = `${fundingSource},${year}`;
                const index = keys.indexOf(fundingSource);
                fundingSourcesPerYear[key][nren] = valuesPerSource[index];
            }
        });
    });
    const fundingSourceDataset = Array.from(Object.entries(fundingSourcesPerYear)).map(([key, data]) => {
        const [fundingSource, year] = key.split(",");
        const color = colorMap.get(fundingSource) || 'black'; // if we get black, we have a problem
        return {
            backgroundColor: color,
            label: fundingSources[fundingSource] + " (" + year + ")",
            data: labelsNREN.map(nren => {
                // ensure that the data is in the same order as the labels
                return data[nren];
            }),
            stack: year,
            borderSkipped: true,
            barPercentage: 0.8,
            borderWidth: 0.5,
            categoryPercentage: 0.8,
            hidden: false,
            datalabels: {
                display: fundingSource == keys[0], // workaround for year labels being duplicated
                formatter: function (_value, context) {
                    return context.dataset.stack;
                },
                font: {
                    family: `"Open Sans", sans-serif`,
                    size: 16,
                    weight: '700'
                },
                anchor: 'start',
                align: 'end',
                offset: function (context) {
                    return context.chart.chartArea.width;
                }
            }
        }
    });

    const dataResponse: BarChartDataset = {
        datasets: fundingSourceDataset,
        labels: labelsNREN.map(l => l.toString())
    }
    return dataResponse;
}

export function createLineChartDataset<T extends NrenAndYearDatapoint>(entries: T[], field: keyof T): BasicDataset {
    const labelsYear = [...new Set(entries.map((item) => item.year))].sort();
    const labelsNREN = [...new Set(entries.map((item) => item.nren))].sort();

    const dataLookup = createDataLookup(entries);

    const sets = labelsNREN.map(nren => {
        const randomColor = stringToColour(nren);
        return {
            backgroundColor: randomColor,
            borderColor: randomColor,
            data: labelsYear.map((year) => {
                const nrendata = dataLookup.get(nren);
                if (!nrendata) {
                    return null;
                }
                const yeardata = nrendata.get(year);
                if (!yeardata) {
                    return null;
                }

                const value = yeardata[field] as number | null;
                return value;
            }),
            label: nren,
            hidden: false
        }
    });

    return {
        datasets: sets,
        labels: labelsYear.map(year => year.toString())
    };
}

export const createNRENStaffDataset = (data: NrenStaff[], roles: boolean, selectedYear: number) => {

    let categories;
    if (roles) {
        categories = [
            "Technical FTE",
            "Non-technical FTE"
        ]
    } else {
        categories = [
            "Permanent FTE",
            "Subcontracted FTE"
        ]
    }
    const fields = {
        "Technical FTE": "technical_fte",
        "Non-technical FTE": "non_technical_fte",
        "Permanent FTE": "permanent_fte",
        "Subcontracted FTE": "subcontracted_fte"
    }

    // get the values for the two categories
    const [category1, category2] = categories
    const [category1Field, category2Field] = [fields[category1], fields[category2]]

    function getPercentages(item: NrenStaff) {
        const category1Value = item[category1Field];
        const category2Value = item[category2Field];
        // calculate the percentages
        const total = category1Value + category2Value
        const category1_percentage = ((category1Value / total) || 0) * 100;
        const category2_percentage = ((category2Value / total) || 0) * 100;
        // round to 2 decimal places
        const result = {};
        result[category1] = Math.round(Math.floor(category1_percentage * 100)) / 100;
        result[category2] = Math.round(Math.floor(category2_percentage * 100)) / 100;
        return result;
    }

    const dataLookup = createDataLookup(data)

    const labelsYear = [selectedYear].sort();
    const labelsNREN = [...new Set(data.map((item: NrenStaff) => item.nren))].sort((nrenA, nrenB) => {
        return nrenA.localeCompare(nrenB);

        // the below sorts by percentages (category1, category2)
        /*
        const categoryMapNrenA = dataLookup.get(nrenA)!.get(selectedYear);
        const categoryMapNrenB = dataLookup.get(nrenB)!.get(selectedYear);
        if (categoryMapNrenA && categoryMapNrenB) {
            const percentagesA = getPercentages(categoryMapNrenA);
            const percentagesB = getPercentages(categoryMapNrenB);

            if (percentagesA[category1] === percentagesB[category1]) {
                return percentagesB[category2] - percentagesA[category1];
            }
            return percentagesB[category1] - percentagesA[category1];
        } else {
            // put NRENs with no data at the end
            if (categoryMapNrenA) {
                return -1
            }
            if (categoryMapNrenB) {
                return 1
            }
            return 0
        }
        */
    });

    const categoriesPerYear = cartesianProduct(categories, labelsYear)

    const nrenStaffDataset = categoriesPerYear.map(function ([category, year]) {

        let color = ""
        if (category === "Technical FTE") {
            color = 'rgba(40, 40, 250, 0.8)'
        } else if (category === "Permanent FTE") {
            color = 'rgba(159, 129, 235, 1)'
        } else if (category === "Subcontracted FTE") {
            color = 'rgba(173, 216, 229, 1)'
        } else if (category === "Non-technical FTE") {
            color = 'rgba(116, 216, 242, 0.54)'
        }

        return {
            backgroundColor: color,
            label: `${category} (${year})`,
            data: labelsNREN.map(nren => {
                // ensure that the data is in the same order as the labels
                const dataForNren = dataLookup.get(nren)!.get(year);
                if (!dataForNren) {
                    return 0
                }
                const percentages = getPercentages(dataForNren);
                return percentages[category];
            }),
            stack: year,
            borderRadius: 10,
            borderSkipped: true,
            barPercentage: 0.8,
            borderWidth: 0.5,
            categoryPercentage: 0.8,
            hidden: false
        }
    })

    const dataset: BarChartDataset = {
        datasets: nrenStaffDataset,
        labels: labelsNREN
    }
    return dataset;
}

export const createTrafficRatioDataset = (data: TrafficRatio[], selectedYear: number) => {

    const categories = [
        "Research & Education",
        "Commodity"
    ]
    const fields = {
        "Research & Education": "r_and_e_percentage",
        "Commodity": "commodity_percentage"
    }

    const dataLookup = createDataLookup(data)

    const labelsYear = [selectedYear].sort();
    const labelsNREN = [...new Set(data.map((item: TrafficRatio) => item.nren))].sort((nrenA, nrenB) => {
        return nrenA.localeCompare(nrenB);

        // the below sorts by percentages (category1, category2)
        /*
        const categoryMapNrenA = dataLookup.get(nrenA)!.get(selectedYear);
        const categoryMapNrenB = dataLookup.get(nrenB)!.get(selectedYear);
        if (categoryMapNrenA && categoryMapNrenB) {
            const percentagesA = getPercentages(categoryMapNrenA);
            const percentagesB = getPercentages(categoryMapNrenB);

            if (percentagesA[category1] === percentagesB[category1]) {
                return percentagesB[category2] - percentagesA[category1];
            }
            return percentagesB[category1] - percentagesA[category1];
        } else {
            // put NRENs with no data at the end
            if (categoryMapNrenA) {
                return -1
            }
            if (categoryMapNrenB) {
                return 1
            }
            return 0
        }
        */
    });

    const categoriesPerYear = cartesianProduct(categories, labelsYear)

    const trafficRatioDataset = categoriesPerYear.map(function ([category, year]) {

        let color = ""
        if (category === "Research & Education") {
            color = 'rgba(40, 40, 250, 0.8)'
        } else if (category === "Commodity") {
            color = 'rgba(116, 216, 242, 0.54)'
        }

        return {
            backgroundColor: color,
            label: `${category} (${year})`,
            data: labelsNREN.map(nren => {
                // ensure that the data is in the same order as the labels
                const dataForNren = dataLookup.get(nren)!.get(year);
                if (!dataForNren) {
                    return 0
                }
                return dataForNren[fields[category]]
            }),
            stack: year,
            borderRadius: 10,
            borderSkipped: true,
            barPercentage: 0.8,
            borderWidth: 0.5,
            categoryPercentage: 0.8,
            hidden: false
        }
    })

    const dataset: BarChartDataset = {
        datasets: trafficRatioDataset,
        labels: labelsNREN
    }

    return dataset;
}

export const createNRENStaffDatasetAbsolute = (data: NrenStaff[], selectedYears: number[]) => {

    const categories = [
        "Permanent FTE",
        "Subcontracted FTE"
    ]

    const fields = {
        "Technical FTE": "technical_fte",
        "Non-technical FTE": "non_technical_fte",
        "Permanent FTE": "permanent_fte",
        "Subcontracted FTE": "subcontracted_fte"
    }

    // get the values for the two categories
    const [category1, category2] = categories
    const [category1Field, category2Field] = [fields[category1], fields[category2]]

    const dataLookup = createDataLookup(data)

    const labelsNREN = [...new Set(data.map((item: NrenStaff) => item.nren))].sort((nrenA, nrenB) => {
        return nrenA.localeCompare(nrenB);
    });

    function getDataset(year, _index) {
        const red = "rgba(219, 42, 76, 1)"

        return {
            backgroundColor: red, //backgroundColors,
            label: `Number of FTEs (${year})`,
            data: labelsNREN.map(nren => {
                // ensure that the data is in the same order as the labels
                const dataForNren = dataLookup.get(nren)!.get(year);
                if (!dataForNren) {
                    return 0
                }
                return (dataForNren[category1Field] ?? 0) + (dataForNren[category2Field] ?? 0);

            }),
            stack: `${year}`,
            borderRadius: 10,
            borderSkipped: true,
            barPercentage: 0.8,
            borderWidth: 0.5,
            categoryPercentage: 0.8,
            hidden: false,
            datalabels: {
                display: true,
                formatter: function (_value, context) {
                    return context.dataset.stack;
                },
                font: {
                    family: `"Open Sans", sans-serif`,
                    size: 16,
                    weight: '700'
                },
                anchor: 'start',
                align: 'end',
                offset: function (context) {
                    return context.chart.chartArea.width;
                }
            }
        }
    }
    const nrenStaffDataset = selectedYears.sort().map(getDataset)

    const dataset: BarChartDataset = {
        datasets: nrenStaffDataset,
        labels: labelsNREN
    }

    return dataset;
}

export const createBarChartDataset = <T extends NrenAndYearDatapoint>(data: T[], columnProperty: keyof T, tooltipPrefix?: string) => {

    const dataLookup = createDataLookup(data)

    const labelsNREN = [...new Set(data.map((item: T) => item.nren))].sort((nrenA, nrenB) => {
        return nrenA.localeCompare(nrenB);
    });
    const labelsYear = [...new Set(data.map((item) => item.year))].sort();

    function getDataset(year, _index) {
        const red = "rgba(219, 42, 76, 1)"

        return {
            backgroundColor: red, //backgroundColors,
            label: `${tooltipPrefix} (${year})`,
            data: labelsNREN.map(nren => {
                // ensure that the data is in the same order as the labels
                const dataForNren = dataLookup.get(nren)!.get(year)!;
                if (!dataForNren) {
                    return 0
                }
                return (dataForNren[columnProperty] as number ?? 0);

            }),
            stack: `${year}`,
            borderRadius: 10,
            borderSkipped: true,
            barPercentage: 0.8,
            borderWidth: 0.5,
            categoryPercentage: 0.8,
            hidden: false,
            datalabels: {
                display: true,
                formatter: function (_value, context) {
                    return context.dataset.stack;
                },
                font: {
                    family: `"Open Sans", sans-serif`,
                    size: 16,
                    weight: '700'
                },
                anchor: 'start',
                align: 'end',
                offset: function (context) {
                    return context.chart.chartArea.width;
                }
            }
        }
    }

    const dataset: BarChartDataset = {
        datasets: labelsYear.sort().map(getDataset),
        labels: labelsNREN
    }

    return dataset;
}
