import { useContext } from 'react';
import { Bar } from 'react-chartjs-2';
import { Chart as ChartJS, CategoryScale, LinearScale, BarElement, Title, Tooltip, Legend } from 'chart.js';

import { NonRAndEPeers } from "compendium/Schema";
import { createBarChartDataset } from "compendium/helpers/dataconversion";
import DataPage from 'compendium/components/DataPage';
import Filter from "compendium/components/graphing/Filter"
import { Sections } from "compendium/helpers/constants";
import { FilterSelectionContext } from 'compendium/providers/FilterSelectionProvider';
import { useData } from 'compendium/helpers/useData';
import ChartContainer from 'compendium/components/graphing/ChartContainer';
import ChartDataLabels from 'chartjs-plugin-datalabels';
import { getBarChartOptions } from 'compendium/helpers/charthelpers';
import titles from "compendium/titles";

ChartJS.register(
    CategoryScale,
    LinearScale,
    BarElement,
    Title,
    Tooltip,
    Legend
);

function NonRAndEPeerPage() {
    const { filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
    const { data, years, nrens } = useData<NonRAndEPeers>('/api/non-re-peers', setFilterSelection);

    const selectedData = data.filter(data =>
        filterSelection.selectedYears.includes(data.year) && filterSelection.selectedNrens.includes(data.nren)
    );
    const datasets = createBarChartDataset(selectedData, 'nr_of_non_r_and_e_peers', 'Number of Peers');

    const filterNode = <Filter
        filterOptions={{ availableYears: [...years], availableNrens: [...nrens.values()] }}
        filterSelection={filterSelection}
        setFilterSelection={setFilterSelection}
    />

    const numNrens = Array.from(new Set(selectedData.map(c => c.nren))).length;
    const heightPerBar = 1.5; // every added bar should give this much additional height

    // set a minimum height of 50rem
    const height = Math.max(numNrens * filterSelection.selectedYears.length * heightPerBar + 5, 50);

    const title = titles['non-rne-peers'];
    const description = `The graph below shows the number of non-Research and Education networks 
    NRENs peer with. This includes all direct IP-peerings to commercial networks, eg. Google`

    const options = getBarChartOptions({
        title: "Number of Non-R&E Peers",
    });

    const filename = "non_r_and_e_peering";
    return (
        <DataPage title={title}
            description={description}
            category={Sections.Network} filter={filterNode}
            data={selectedData} filename={filename}>

            <ChartContainer>
                <div className="chart-container" style={{ 'height': `${height}rem` }}>
                    <Bar
                        data={datasets}
                        options={options}
                        plugins={[ChartDataLabels]}
                    />
                </div>
            </ChartContainer>
        </DataPage>
    );
}
export default NonRAndEPeerPage;
