import { ReactElement, useContext } from 'react';
import { Line } from 'react-chartjs-2';

import {
    Chart as ChartJS,
    CategoryScale,
    LinearScale,
    PointElement,
    LineElement,
    Title,
    Tooltip,
    Legend,
} from 'chart.js';

import { Budget } from "compendium/Schema";
import { createLineChartDataset } from "compendium/helpers/dataconversion";
import DataPage from 'compendium/components/DataPage';
import Filter from "compendium/components/graphing/Filter";
import { Sections } from "compendium/helpers/constants";
import { FilterSelectionContext } from 'compendium/providers/FilterSelectionProvider';
import ChartContainer from "compendium/components/graphing/ChartContainer";
import { useData } from 'compendium/helpers/useData';
import { getLineChartOptions } from 'compendium/helpers/charthelpers';
import titles from "compendium/titles";

ChartJS.register(
    CategoryScale,
    LinearScale,
    PointElement,
    LineElement,
    Title,
    Tooltip,
    Legend
);

function BudgetPage(): ReactElement {

    const field = 'budget';
    const { filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
    const { data, nrens } = useData<Budget>('/api/budget', setFilterSelection);

    const selectedData = data.filter(data =>
        filterSelection.selectedNrens.includes(data.nren) // we only allow filtering nrens for this page
    );
    const budgetData = createLineChartDataset(selectedData, field);

    const filterNode = <Filter
        filterOptions={{ availableYears: [], availableNrens: [...nrens.values()] }}
        filterSelection={filterSelection}
        setFilterSelection={setFilterSelection}
    />

    const options = getLineChartOptions({ title: 'Budget in M€', tooltipUnit: 'M€', unit: 'M€' });

    return (
        <DataPage title={titles.budget}
            description={<span>The graph shows NREN budgets per year (in millions Euro). When budgets are not per calendar year, the NREN is asked to provide figures of the budget that covers the largest part of the year, and to include any GÉANT subsidy they may receive.
                <br />NRENs are free to decide how they define the part of their organisation dedicated to core NREN business, and the budget. The merging of different parts of a large NREN into a single organisation, with a single budget can lead to significant changes between years, as can receiving funding for specific time-bound projects.
                <br />Hovering over the graph data points shows the NREN budget for the year. Gaps indicate that the budget question was not filled in for a particular year.</span>} category={Sections.Organisation} filter={filterNode}
            data={selectedData} filename="budget_data">
            <ChartContainer>
                <Line data={budgetData} options={options} />
            </ChartContainer>
        </DataPage>
    );
}

export default BudgetPage;
