import { useContext } from 'react';
import { Bar } from 'react-chartjs-2';
import { Col, Row } from "react-bootstrap";
import { Chart as ChartJS, CategoryScale, LinearScale, BarElement, Title, Tooltip, Legend } from 'chart.js';
import ChartDataLabels from 'chartjs-plugin-datalabels';

import { FundingSource } from "compendium/Schema";
import { createFundingSourceDataset } from "compendium/helpers/dataconversion";
import DataPage from 'compendium/components/DataPage';
import Filter from "compendium/components/graphing/Filter"
import { Sections } from "compendium/helpers/constants";
import ColorBadge from 'compendium/components/ColorBadge';
import { FilterSelectionContext } from 'compendium/providers/FilterSelectionProvider';
import ChartContainer from "compendium/components/graphing/ChartContainer";
import { useData } from 'compendium/helpers/useData';
import titles from "compendium/titles";

ChartJS.register(
  CategoryScale,
  LinearScale,
  BarElement,
  Title,
  Tooltip,
  Legend
);

const chartOptions = {
  maintainAspectRatio: false,
  layout: {
    padding: {
      right: 60
    }
  },
  animation: {
    duration: 0,
  },
  plugins: {
    legend: {
      display: false
    },
    tooltip: {
      callbacks: {
        label: function (tooltipItem) {
          let label = tooltipItem.dataset.label || '';

          if (tooltipItem.parsed.y !== null) {
            label += `: ${tooltipItem.parsed.x}%`
          }
          return label;
        }
      },
    },
  },
  scales: {
    x: {
      position: "top" as const,
      ticks: {
        callback: (value: string | number) => {
          return `${value}%`;
        },
        stepSize: 10,
      },
      max: 100,
      min: 0
    },
    xBottom: {
      ticks: {
        callback: (value: string | number) => {
          return `${value}%`;
        },
        stepSize: 10,
      },
      max: 100,
      min: 0,
      grid: {
        drawOnChartArea: false
      },
      afterDataLimits: function (axis) {
        const indices = Object.keys(ChartJS.instances)

        // initial values should be far outside possible range
        let max = -999999
        let min = 999999

        for (const index of indices) {
          if (ChartJS.instances[index] && axis.chart.scales.xBottom) {
            min = Math.min(ChartJS.instances[index].scales.x.min, min);
            max = Math.max(ChartJS.instances[index].scales.x.max, max);
          }
        }

        axis.chart.scales.xBottom.options.min = min;
        axis.chart.scales.xBottom.options.max = max;
        axis.chart.scales.xBottom.min = min;
        axis.chart.scales.xBottom.max = max;
      },
    },
    y: {
      ticks: {
        autoSkip: false
      },
    }
  },
  indexAxis: "y" as const,
};

function FundingSourceLegend() {
  return (
    <div className="d-flex justify-content-center bold-grey-12pt">
      <Row xs="auto" className="border rounded-3 border-1 my-5 justify-content-center">
        <Col className="d-flex align-items-center">
          <ColorBadge key={0} index={0} />Client Institutions
        </Col>
        <Col className="d-flex align-items-center">
          <ColorBadge key={1} index={1} />Commercial
        </Col>
        <Col className="d-flex align-items-center">
          <ColorBadge key={2} index={2} />European Funding
        </Col>
        <Col className="d-flex align-items-center">
          <ColorBadge key={3} index={3} />Gov/Public Bodies
        </Col>
        <Col className="d-flex align-items-center">
          <ColorBadge key={4} index={4} />Other
        </Col>
      </Row>
    </div>
  );
}

function FundingSourcePage() {
  const { filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
  const { data: fundingSourceData, years, nrens } = useData<FundingSource>('/api/funding', setFilterSelection);

  const selectedData = fundingSourceData.filter(data =>
    filterSelection.selectedYears.includes(data.year) && filterSelection.selectedNrens.includes(data.nren)
  );

  const fundingSourceDataset = createFundingSourceDataset(selectedData);

  // remove the datapoints and labels for the nrens that aren't selected
  // unfortunately we cannot just hide them because graph.js doesn't want
  // to create a stack from a single dataset
  fundingSourceDataset.datasets.forEach(dataset => {
    dataset.data = dataset.data.filter((e, i) => {
      return filterSelection.selectedNrens.includes(fundingSourceDataset.labels[i]);
    });
  });
  fundingSourceDataset.labels = fundingSourceDataset.labels.filter((e) => filterSelection.selectedNrens.includes(e));

  const filterNode = <Filter
    filterOptions={{ availableYears: [...years], availableNrens: [...nrens.values()] }}
    filterSelection={filterSelection}
    setFilterSelection={setFilterSelection}
  />

  const numNrens = Array.from(new Set(selectedData.map(c => c.nren))).length;
  const numYears = filterSelection.selectedYears.length;
  const heightPerBar = 2; // every added bar should give this much additional height

  // set a minimum height of 20rem, additional years need some more space
  const height = numNrens * numYears * heightPerBar + 5;
  return (
    <DataPage title={titles.funding}
      description={<span>The graph shows the percentage share of their income that NRENs derive from different sources, with any funding and NREN may receive from GÉANT included within &quot;European funding&quot;. By &quot;Client institutions&quot; NRENs may be referring to universities, schools, research institutes, commercial clients, or other types of organisation. &quot;Commercial services&quot; include services such as being a domain registry, or security support.
        <br />Hovering over the graph bars will show the exact figures, per source. When viewing multiple years, it is advisable to restrict the number of NRENs being compared.</span>}
      category={Sections.Organisation} filter={filterNode}
      data={selectedData} filename='income_source_of_nren_per_year'>

      <ChartContainer>
        <FundingSourceLegend />
        <div className="chart-container" style={{ 'height': `${height}rem` }}>
          <Bar
            plugins={[ChartDataLabels]}
            data={fundingSourceDataset}
            options={chartOptions}
          />
        </div>
        <FundingSourceLegend />
      </ChartContainer>
    </DataPage>
  );
}
export default FundingSourcePage;
