import { useContext } from 'react';

import {
    BarElement,
    CategoryScale,
    Chart as ChartJS,
    Legend,
    LinearScale,
    Title,
    Tooltip,
} from 'chart.js';
import ChartDataLabels from 'chartjs-plugin-datalabels';
import { Bar } from 'react-chartjs-2';

import DataPage from 'compendium/components/DataPage';
import ChartContainer from 'compendium/components/graphing/ChartContainer';
import Filter from 'compendium/components/graphing/Filter';
import { getBarChartOptions } from 'compendium/helpers/charthelpers';
import { Sections } from "compendium/helpers/constants";
import { createBarChartDataset } from 'compendium/helpers/dataconversion';
import { useData } from 'compendium/helpers/useData';
import { FilterSelectionContext } from 'compendium/providers/FilterSelectionProvider';
import { CentralProcurement } from "compendium/Schema";
import titles from "compendium/titles";

ChartJS.register(
    CategoryScale,
    LinearScale,
    BarElement,
    Title,
    Tooltip,
    Legend
);

function CentralProcurementPage() {
    const field = 'amount';
    const validityCheck = (data: CentralProcurement) => data[field] != null;
    const { filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
    const { data, years, nrens } = useData<CentralProcurement>('/api/central-procurement', setFilterSelection, validityCheck);

    const selectedData = data.filter(data =>
        filterSelection.selectedYears.includes(data.year) && filterSelection.selectedNrens.includes(data.nren)
    );
    const datasets = createBarChartDataset(selectedData, field, 'Procurement Value');

    const filterNode = <Filter
        filterOptions={{ availableYears: [...years], availableNrens: [...nrens.values()] }}
        filterSelection={filterSelection}
        setFilterSelection={setFilterSelection}
    />

    const numNrens = Array.from(new Set(selectedData.map(c => c.nren))).length;
    const heightPerBar = 1.5; // every added bar should give this much additional height

    // set a minimum height of 50rem
    const height = Math.max(numNrens * filterSelection.selectedYears.length * heightPerBar + 5, 50);

    const title = titles["central-procurement"];
    const description = <span>Some NRENs centrally procure software for their customers.
        The graph below shows the total value (in Euro) of software procured in the previous year by the NRENs.
        Please note you can only see the select NRENs which carry out this type of procurement.
        Those who do not offer this are not selectable.</span>;

    const options = getBarChartOptions({
        title: "Software Procurement Value",
        valueTransform(value) {
            const formatter = new Intl.NumberFormat(void 0, {
                style: 'currency',
                currency: 'EUR',
                trailingZeroDisplay: 'stripIfInteger',
            });
            return `${formatter.format(value)}`
        },
    });

    const filename = "central_procurement";
    return (
        <DataPage title={title}
            description={description}
            category={Sections.Policy} filter={filterNode}
            data={selectedData} filename={filename}>

            <ChartContainer>
                <div className="chart-container" style={{ 'height': `${height}rem` }}>
                    <Bar
                        data={datasets}
                        options={options}
                        plugins={[ChartDataLabels]}
                    />
                </div>
            </ChartContainer>
        </DataPage>
    );
}
export default CentralProcurementPage;
