import { useContext } from "react";
import ProgressBar from 'compendium/survey/ProgressBar';
import { Container, Row } from "react-bootstrap";
import { userContext } from "compendium/providers/UserProvider";

function SurveyNavigationComponent({ surveyModelContainer, surveyActions, year, nren, children, onPageChange
 }) {
    const { surveyModel } = surveyModelContainer;
    const { user: loggedInUser } = useContext(userContext);
    const pageNo = surveyModel?.currentPageNo ?? 0;
    const editing = surveyModel?.mode === 'edit';
    const lockedBy = surveyModel?.["lockedBy"] ?? "";
    const responseStatus = surveyModel?.["status"] ?? "";

    const incrementPageNo = () => { onPageChange(surveyModel.currentPageNo + 1); };

    const doSurveyAction = async (action) => {
        await surveyActions[action]();
    }

    const renderButton = (text, action) => {
        return (
            <button className="sv-btn sv-btn--navigation" onClick={action}>
                {text}
            </button>
        );
    };

    const renderActionButton = (text, action) => {
        return renderButton(text, () => doSurveyAction(action));
    };

    const saveAndStopEdit = 'Save and stop editing';
    const save = 'Save progress';
    const startEditing = 'Start editing';
    const completeSurvey = 'Complete Survey';

    const renderExternalNavigation = () => {
        return (
            <div className="survey-edit-buttons-block">
                {!editing && !lockedBy && surveyModel.editAllowed && renderActionButton(startEditing, 'startEdit')}
                {!editing && lockedBy && lockedBy == loggedInUser.name && renderActionButton('Discard any unsaved changes and release your lock', 'releaseLock')}
                {editing && renderActionButton(save, 'save')}
                {editing && renderActionButton(saveAndStopEdit, 'saveAndStopEdit')}
                {editing && renderActionButton(completeSurvey, 'complete')}
                {(pageNo !== surveyModel.visiblePages.length - 1) && renderButton('Next Section', incrementPageNo)}
            </div>
        );
    };

    return (
        <Container>
            <Row className="survey-content">
                <h2><span className="survey-title">{year} Compendium Survey </span><span className="survey-title-nren"> {nren} </span><span> - {responseStatus}</span></h2>

                <div style={{ marginTop: '1rem', textAlign: 'justify' }}>
                    <p>
                        To get started, click “{startEditing}” to end read-only mode. Different people from your NREN (Compendium administrators) can contribute to the survey if needed, but agreement should be reached internally before completing the survey as the administration team will treat responses as a single source of truth from the NREN.
                        You can start editing only when nobody else from your NREN is currently working on the survey.</p>
                    <p><b>In a small change, the survey now asks about this calendar year, i.e. {year}</b> (or the current financial year if your budget or staffing data does not match the calendar year). For network questions, please provide data from the 12 months preceding you answering the question.
                        Where available, the survey questions are pre-filled with answers from the previous survey.
                        You can edit the pre-filled answer to provide new information, or press the “no change from previous year” button.</p>
                    <p>Press the “{save}“ or “{saveAndStopEdit}“ button to save all answers in the survey. When you reach the last section of the survey (Services), you will find a “{completeSurvey}“ button which saves all answers in the survey and lets the Compendium team know that your answers are ready to be published.
                        As long as the survey remains open, any Compendium administrator from your NREN can add answers or amend existing ones, even after using the “{completeSurvey}“ button.</p>
                    <p>Some fields require specific data, such as numerical data, valid http-addresses, and in some questions, the answer has to add up to 100%. If an answer does not fulfil the set criteria, the question will turn pink and an error message will appear. Fields can be left blank if you prefer not to answer a question.
                        If you notice any errors after the survey was closed, please contact us for correcting those.</p>
                </div>
                <p>Thank you for taking the time to fill in the {year} Compendium Survey. Any questions or requests can be sent to <a href={`mailto:Partner-Relations@geant.org`}><span>Partner-Relations@geant.org</span></a></p>
                {editing && <><br /><b>Remember to click “{saveAndStopEdit}” before leaving the page.</b></>}
            </Row>
            <Row>
                {renderExternalNavigation()}
            </Row>
            <Row className="survey-content">
                {!editing && (
                    <div className="survey-edit-explainer">
                        {!lockedBy && surveyModel.editAllowed && 'The survey is in read-only mode; click the “Start editing“ button to begin editing the answers.'}
                        {!lockedBy && !surveyModel.editAllowed && 'The survey is in read-only mode and can not be edited by you.'}
                        {lockedBy && lockedBy != loggedInUser.name && 'The survey is in read-only mode and currently being edited by: ' + lockedBy + '. To start editing the survey, ask them to complete their edits.'}
                        {lockedBy && lockedBy == loggedInUser.name && 'The survey is in read-only mode because you started editing in another tab, browser or device. To start editing the survey, either complete those edits or click the "Discard any unsaved changes" button.'}
                    </div>
                )}
            </Row>
            <Row>
                <ProgressBar surveyModel={surveyModel} pageNoSetter={onPageChange} pageNo={pageNo} />
                {children}
            </Row>
            <Row>
                {renderExternalNavigation()}
            </Row>

        </Container>
    );
}

export default SurveyNavigationComponent;
