import path from "path";
import fs from "fs";
import ForkTsCheckerWebpackPlugin from "fork-ts-checker-webpack-plugin";
import MiniCssExtractPlugin from "mini-css-extract-plugin";

import { Configuration as WebpackConfiguration } from "webpack";
import { Configuration as WebpackDevServerConfiguration } from "webpack-dev-server";

interface Configuration extends WebpackConfiguration {
  devServer?: WebpackDevServerConfiguration;
}

const isProduction = process.argv[process.argv.indexOf('--mode') + 1] === 'production';

console.log(isProduction ? 'Production build' : 'Development build')

const config: Configuration = {
  entry: "./src/index.tsx",
  module: {
    rules: [
      {
        test: /\.(ts|js)x?$/,
        exclude: /node_modules/,
        use: {
          loader: "babel-loader",
          options: {
            presets: [
              "@babel/preset-env",
              "@babel/preset-react",
              "@babel/preset-typescript",
            ],
          },
        },
      },
      {
        test: /\.scss$/,
        exclude: /node_modules/,
        use: [
          isProduction ? MiniCssExtractPlugin.loader : "style-loader",
          "css-loader",
          "sass-loader",
        ],
      },
      {
        test: /\.css$/,
        use: [
          isProduction ? MiniCssExtractPlugin.loader : "style-loader",
          "css-loader"
        ],
      },
      {
        test: /\.(png|svg|jpe?g|gif)$/,
        type: 'asset/resource'
      },
    ],
  },
  resolve: {
    extensions: [".tsx", ".ts", ".js", ".html", '.svg'],
  },
  output: {
    path: path.resolve(__dirname, "static"),
    filename: "bundle.js",
  },
  devServer: {
    static: path.join(__dirname, "static"),
    compress: true,
    port: 54054,
    // Allow SPA urls to work with dev-server
    historyApiFallback: true,
    setupMiddlewares: (middlewares, devServer) => {
      if (!devServer) {
        throw new Error('webpack-dev-server is not defined');
      }

      const getLicense = (req, res) => {
        const licensePath = path.resolve(__dirname, 'survey-creator.license');
        if (fs.existsSync(licensePath)) {
          const contents = fs.readFileSync(licensePath, 'utf8');
          res.send(contents);
        } else {
          res.status(404).send('License key not found');
        }
      }

      const getSurveyJson = (req, res) => {
        const surveyModel = path.resolve(__dirname, 'models', 'survey_model.json');
        if (fs.existsSync(surveyModel)) {
          const contents = fs.readFileSync(surveyModel, 'utf8');
          res.send(contents);
        } else {
          res.send(JSON.stringify({}));
        }
      }

      devServer.app?.use('/license-key', getLicense);
      devServer.app?.use('/survey-json', getSurveyJson);

      return middlewares;
    },
  },
  devtool: isProduction ? undefined : "inline-source-map",
  plugins: [
    new ForkTsCheckerWebpackPlugin({
      async: false,
    }),
    new MiniCssExtractPlugin(
      {
        filename: 'bundle.css',
      }
    ),
  ],
};

export default config;
