import logging
from typing import Any, Union

from flask import Blueprint, jsonify, request
from flask_login import current_user, AnonymousUserMixin, login_required  # type: ignore
from sqlalchemy import select

from compendium_v2.auth.session_management import admin_required
from compendium_v2.db import db
from compendium_v2.db.auth_model import User, ROLES
from compendium_v2.db.presentation_models import NREN
from compendium_v2.routes import common

routes = Blueprint('user', __name__)
logger = logging.getLogger(__name__)

USER_RESPONSE_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',

    'definitions': {
        'permissions': {
            'type': 'object',
            'properties': {
                'admin': {'type': 'boolean'},
                'active': {'type': 'boolean'},
            },
            'required': ['admin', 'active'],
            'additionalProperties': False
        },
        'user': {
            'type': 'object',
            'properties': {
                'name': {'type': ['string', 'null']},
                'email': {'type': ['string', 'null']},
                'permissions': {'$ref': '#/definitions/permissions'},
                'id': {'type': 'uuid'},
                'role': {'type': 'string'},
                'oidc_sub': {'type': 'string'},
                'nrens': {
                    'type': 'array',
                    'items': {'type': 'string'}
                }
            },
            'required': ['permissions'],
            'additionalProperties': False
        }
    },

    'type': 'array',
    'items': {'$ref': '#/definitions/charging'}
}


def _extract_user(user: Union[User, AnonymousUserMixin]):
    if isinstance(user, AnonymousUserMixin):
        return {
            'permissions': {
                'admin': False,
                'active': False,
            }
        }
    return {
        'name': user.fullname,
        'email': user.email,
        'permissions': {
            'admin': user.roles == ROLES.admin,
            'active': user.active,
        },
        'id': user.id,
        'role': user.roles.value,
        'oidc_sub': user.oidc_sub,
        'nrens': [nren.name for nren in user.nrens]
    }


@routes.route('/', methods=['GET'])
@common.require_accepts_json
def current_user_view() -> Any:
    """
    handler for /api/user/ requests

    response will be formatted as:

    .. asjson::
        compendium_v2.routes.user.USER_RESPONSE_SCHEMA

    :return:
    """

    return jsonify(_extract_user(current_user))


@routes.route('/list', methods=['GET'])
@common.require_accepts_json
@login_required
@admin_required
def all_users_view() -> Any:
    # TODO schema and docstring

    entries = [_extract_user(user) for user in db.session.scalars(
        select(User).order_by(User.email)).unique()]
    return jsonify(entries)


@routes.route('/<user_id>', methods=['PUT'])
@common.require_accepts_json
@login_required
@admin_required
def update_user_view(user_id) -> Any:
    """
    Handler for /api/user/<user_id> via PUT request.

    Request data should be in JSON format with the fields you want to update.
    The response will be formatted the same way as in the current_user_view.

    Example JSON request data::

        {
            "role": "admin",
            "active": False,
            "nren": "1"
        }

    :return:
    """

    def _update_user_data(user: User, update_data: dict):

        new_role = update_data.get('role', user.roles.value)

        if 'role' in update_data:
            if new_role != user.roles.value:
                if user == current_user:
                    return jsonify({'success': False, 'message': 'Cannot change your own role.'}), 400

                try:
                    new_role = ROLES(new_role)
                except ValueError:
                    return jsonify({'success': False, 'message': 'Invalid role.'}), 400
                logger.info(f'Updating role for user {user.id} from {user.roles} to {new_role}')
                user.roles = new_role

        elif 'active' in update_data:
            _active = bool(update_data['active'])
            if _active != user.active:
                if user == current_user:
                    return jsonify({'success': False, 'message': 'Cannot deactivate yourself.'}), 400

                user.active = _active
                logger.info(f'Updating active status for user {user.id} to {_active}')

        elif 'nren' in update_data:
            new_nrens = None
            new_id = update_data['nren']
            if not new_id:
                # if no ID is provided, remove all NRENs
                user.nrens = []
            else:
                try:
                    new_nrens = db.session.scalars(select(NREN).filter(NREN.id == update_data['nren'])).all()
                    logger.info(f'Updating NRENs for user {user.id} to {new_nrens}')
                    user.nrens = [nren for nren in new_nrens]
                except Exception:
                    return jsonify({'success': False, 'message': 'Error when updating user NREN'}), 500

        db.session.commit()
        return jsonify({'success': True, 'message': 'User updated successfully', 'user': _extract_user(user)})

    body = request.get_json()
    if not body:
        return jsonify({"success": False, 'message': 'Invalid request'}), 400

    if not user_id:
        return jsonify({"success": False, 'message': 'Invalid user ID'}), 400

    user = db.session.execute(select(User).filter_by(id=user_id)).scalar()

    if not user:
        return jsonify({"success": False, 'message': 'User not found.'}), 404

    return _update_user_data(user, body)


@routes.route('/<user_id>', methods=['DELETE'])
@common.require_accepts_json
@login_required
@admin_required
def delete_user_view(user_id) -> Any:
    """
    Handler for /api/user/<user_id> via DELETE request.

    Deletes a user from the database, given a user ID.
    User cannot delete themselves.

    :param user_id: the user ID to delete
    :return:
    """

    user = db.session.execute(select(User).filter_by(id=user_id)).scalar()

    if not user:
        return jsonify({"success": False, 'message': 'User not found.'}), 404

    if user == current_user:
        return jsonify({'success': False, 'message': 'Cannot delete yourself.'}), 400

    db.session.delete(user)
    db.session.commit()

    return jsonify({'success': True, 'message': f'User {user.email} deleted successfully'})
