import csv
import os
import json
import compendium_v2
import pytest
from compendium_v2.auth.session_management import setup_login_manager, User, ROLES
from compendium_v2.db import db, presentation_models, survey_models
from compendium_v2.survey_db import model as survey_db_model
from flask_login import LoginManager  # type: ignore
from sqlalchemy import select


def _test_data(filename, as_csv=True):
    data_filename = os.path.join(os.path.dirname(__file__), 'data', filename)

    if not as_csv:
        return open(data_filename, "r")

    return csv.DictReader(open(data_filename, "r"))


@pytest.fixture
def dummy_config():
    yield {
        'SQLALCHEMY_DATABASE_URI': 'sqlite://',
        'SURVEY_DATABASE_URI': 'sqlite:///',
        'SECRET_KEY': 'testsecret123'
    }


@pytest.fixture
def nren_services():
    def load_services(app):
        with app.app_context():
            with _test_data('presentation_models_data/Service.json', as_csv=False) as f:
                services = json.load(f)
                for row in services:
                    service = presentation_models.Service(
                        **row
                    )
                    db.session.merge(service)
                db.session.commit()
    yield load_services


@pytest.fixture
def mocked_admin_user(app, test_survey_data, mocker):
    with app.app_context():
        user = User(email='testemail123@email.local', fullname='testfullname', oidc_sub='fakesub', roles=ROLES.admin)

        nren2 = db.session.scalar(select(presentation_models.NREN).filter(presentation_models.NREN.name == 'nren2'))
        user.nrens.append(nren2)
        db.session.add(user)
        db.session.commit()

        def user_loader(*args):
            return user

        mocker.patch('flask_login.utils._get_user', user_loader)
        yield user


@pytest.fixture
def mocked_user(app, mocker):
    with app.app_context():
        user = User(email='testemail123@email.local', fullname='testfullname', oidc_sub='fakesub')

        nren2 = db.session.scalar(select(presentation_models.NREN).filter(presentation_models.NREN.name == 'nren2'))
        user.nrens.append(nren2)
        db.session.add(user)
        db.session.commit()

        def user_loader(*args):
            return user

        mocker.patch('flask_login.utils._get_user', user_loader)
        yield user


@pytest.fixture
def mocked_observer_user(app, test_survey_data, mocker):
    with app.app_context():
        user = User(email='observer123@email.local', fullname='observerfullname',
                    oidc_sub='fakesub', roles=ROLES.observer)

        db.session.add(user)

        def user_loader(*args):
            return user

        mocker.patch('flask_login.utils._get_user', user_loader)


@pytest.fixture
def test_survey_data(app):
    with app.app_context():
        nren_names = ['nren1', 'nren2', 'nren3', 'nren4']
        nren_dict = {nren_name: presentation_models.NREN(
            id=idx+9999, name=nren_name, country='country') for idx, nren_name in enumerate(nren_names)}
        db.session.add_all(nren_dict.values())

        survey2021 = survey_models.Survey(year=2021, survey={}, status=survey_models.SurveyStatus.published)
        survey2022 = survey_models.Survey(year=2022, survey={}, status=survey_models.SurveyStatus.published)
        survey2023 = survey_models.Survey(
            year=2023,
            survey={'part1': [{'title': 'ha', 'visibleIf': 'false'}]},
            status=survey_models.SurveyStatus.open
        )
        db.session.add_all([survey2021, survey2022, survey2023])

        db.session.add(survey_models.SurveyResponse(
            nren=nren_dict['nren1'],
            survey=survey2023,
            answers={},
            status=survey_models.ResponseStatus.completed
        ))

        db.session.commit()


@pytest.fixture
def test_survey_data_with_responses(app):

    test_survey = json.load(_test_data('survey_model_2024_testdata.json', False))
    test_response_data = json.load(_test_data('nren2_test_response_data.json', False))
    with app.app_context():
        nren_names = ['nren1', 'nren2', 'nren3', 'nren4']
        nren_dict = {nren_name: presentation_models.NREN(
            id=idx+999, name=nren_name, country='country') for idx, nren_name in enumerate(nren_names)}
        db.session.add_all(nren_dict.values())

        survey2021 = survey_models.Survey(year=2021, survey={}, status=survey_models.SurveyStatus.published)
        survey2022 = survey_models.Survey(year=2022, survey={}, status=survey_models.SurveyStatus.published)
        survey2023 = survey_models.Survey(year=2023, survey={}, status=survey_models.SurveyStatus.published)
        survey2024 = survey_models.Survey(
            year=2024,
            survey=test_survey,
            status=survey_models.SurveyStatus.open
        )
        db.session.add_all([survey2021, survey2022, survey2023, survey2024])

        responses = []

        for name, nren in nren_dict.items():
            response = survey_models.SurveyResponse(
                nren=nren,
                survey=survey2023,
                answers={'data': test_response_data},
                status=survey_models.ResponseStatus.completed
            )
            responses.append(response)

        db.session.add_all(responses)
        db.session.commit()


@pytest.fixture
def app(dummy_config):
    app = compendium_v2._create_app_with_db(dummy_config)
    app.config['LOGIN_DISABLED'] = True  # disable login and admin_required decorators
    login_manager = LoginManager()
    login_manager.init_app(app)
    login_manager.login_view = 'authentication.login'
    setup_login_manager(login_manager)

    with app.app_context():
        db.create_all(bind_key=None)
    yield app


@pytest.fixture
def app_with_survey_db(dummy_config):
    dummy_config['SQLALCHEMY_BINDS'] = {survey_db_model.SURVEY_DB_BIND: dummy_config['SURVEY_DATABASE_URI']}
    app = compendium_v2._create_app_with_db(dummy_config)
    with app.app_context():
        db.create_all()
    yield app


@pytest.fixture
def client(app):
    return app.test_client()
