import os
import openpyxl
import json

from compendium_v2.db import db, presentation_models
from compendium_v2.publishers.legacy_publisher.survey_publisher_legacy_db import FundingSource, \
    StaffQuestion, OrgQuestion, ChargingStructure, ECQuestion
from compendium_v2.db.survey_models import Survey, SurveyStatus, SurveyResponse
from compendium_v2.publishers.legacy_publisher.survey_publisher_legacy import _cli, VALID_KEYS_2024_SURVEY


def _test_data(filename):
    data_filename = os.path.join(os.path.dirname(__file__), 'data', filename)

    with open(data_filename, "r") as f:
        return json.load(f)


def org_data(question):
    """
    This function defines test data for the org questions.

    The following data is defined for the appropriate questions as modeled in Compendium:

    nren1,CYNET-CSIRT,cert team
    nren1,DFN-CERT,CERT
    nren2,Educampus Services,MIS shared services for third level.
    nren3,VilniusTech,Technical centre
    nren3,KU,Technical centre
    nren3,VDU,Technical centre
    nren3,VU,Technical centre
    nren3,KTU,"NOC, administrative authority"
    """

    if question == OrgQuestion.PARENT_ORG_NAME:
        return [
            ('nren1', 'Org1'),
            ('nren3', 'Org3'),
        ]

    if str(question.name).endswith('1_NAME'):
        return [
            ('nren1', 'CYNET-CSIRT'),
            ('nren2', 'Educampus Services'),
            ('nren3', 'VilniusTech'),
        ]

    if str(question.name).endswith('2_NAME'):
        return [
            ('nren1', 'DFN-CERT'),
            ('nren3', 'KU'),
        ]

    if str(question.name).endswith('3_NAME'):
        return [
            ('nren3', 'VDU'),
        ]

    if str(question.name).endswith('4_NAME'):
        return [
            ('nren3', 'VU'),
        ]

    if str(question.name).endswith('5_NAME'):
        return [
            ('nren3', 'KTU'),
        ]

    if str(question.name).endswith('1_CHOICE'):
        return [
            ('nren1', 'other'),
            ('nren2', 'other'),
            ('nren3', 'Technical centre'),
        ]

    if str(question.name).endswith('2_CHOICE'):
        return [
            ('nren1', 'other'),
            ('nren3', 'Technical centre'),
        ]

    if str(question.name).endswith('3_CHOICE'):
        return [
            ('nren3', 'Technical centre'),
        ]

    if str(question.name).endswith('4_CHOICE'):
        return [
            ('nren3', 'Technical centre'),
        ]

    if str(question.name).endswith('5_CHOICE'):
        return [
            ('nren3', 'other'),
        ]

    if str(question.name).endswith('1_ROLE'):
        return [
            ('nren1', 'cert team'),
            ('nren2', 'MIS shared services for third level.')
        ]

    if str(question.name).endswith('2_ROLE'):
        return [
            ('nren1', 'CERT'),
        ]

    if str(question.name).endswith('3_ROLE'):
        return []

    if str(question.name).endswith('4_ROLE'):
        return []

    if str(question.name).endswith('5_ROLE'):
        return [
            ('nren3', 'NOC, administrative authority')
        ]


def setup_survey_db(app_with_survey_db, mocker):

    def get_rows_as_tuples(*args, **kwargs):
        return [
            ('nren1', '100'),
            ('nren2', '200'),
            ('nren3', '300'),
            ('nren4', 'abcd')
        ]

    def funding_source_data():
        yield FundingSource.CLIENT_INSTITUTIONS, [
            ('nren1', '10'),
            ('nren2', '80'),
            ('nren3', '30'),
        ]
        yield FundingSource.EUROPEAN_FUNDING, [
            ('nren1', '50'),
            ('nren2', '20'),
            ('nren3', '30'),
        ]
        yield FundingSource.OTHER, [
            ('nren1', '40'),
            ('nren2', 'abc'),
            ('nren3', '30'),
        ]

    def question_data(question):
        if question == StaffQuestion.NON_TECHNICAL_FTE:
            return [
                ('nren1', '10'),
                ('nren2', '80'),
                ('nren3', '30'),
            ]

        if question == StaffQuestion.TECHNICAL_FTE:
            return [
                ('nren1', '50'),
                ('nren2', '20'),
                ('nren3', '30'),
            ]

        if question == StaffQuestion.PERMANENT_FTE:
            return [
                ('nren1', '60'),
                ('nren2', 'abc'),
                ('nren3', '30'),
            ]

        if question == StaffQuestion.SUBCONTRACTED_FTE:
            return [
                ('nren1', '0'),
                ('nren2', '0'),
                ('nren3', '0'),
            ]
        if question in OrgQuestion:
            return org_data(question)

        if question == ChargingStructure.charging_structure:
            return [
                ('nren1', 'We do not charge them directly'),
                ('nren2', 'We charge a usage-based fee'),
                ('nren3', 'Other'),
            ]

        if question in ECQuestion:
            return [
                ('nren1', '[""'),
                ('nren2', '["project1", "project2"]'),
                ('nren3', '["project3"]'),
            ]

    def question_id_data(question_id, year):
        if question_id in [
                16469, 16064, 15720, 15305, 14910, 16471, 16066, 15722, 15307, 14912, 16473, 16378,
                16475, 16068, 15724, 15309, 14914, 16477, 16070, 15726, 15311, 14916, 16479, 16072, 15728, 15575,
                16481, 16074, 15730, 15577, 16761]:
            return [
                ('nren1', f'www.nren.com/somepolicy{year}.pdf'),
                ('nren2', 'policyemail@nren.com'),
                ('nren3', 'n.a. online'),
            ]

    def institutions_urls_data(question_id):
        if question_id == 16507:
            return [
                (87483, 'ANA', 2013, "http://www.rash.al/index.php/network/points-of-presence-pop"),
                (163286, 'ANA', 2014, "http://www.rash.al/index.php/network/points-of-presence-pop"),
            ]
        else:
            return []

    mocker.patch('compendium_v2.publishers.legacy_publisher.survey_publisher_legacy_db.query_budget',
                 get_rows_as_tuples)
    mocker.patch('compendium_v2.publishers.legacy_publisher.survey_publisher_legacy_db.query_funding_sources',
                 funding_source_data)
    mocker.patch('compendium_v2.publishers.legacy_publisher.survey_publisher_legacy_db.query_question',
                 question_data)
    mocker.patch('compendium_v2.publishers.legacy_publisher.survey_publisher_legacy_db.query_question_id',
                 question_id_data)
    mocker.patch('compendium_v2.publishers.legacy_publisher.survey_publisher_legacy_db.recursive_query',
                 institutions_urls_data)


def test_legacy_publisher(app_with_survey_db, mocker, nren_services):
    setup_survey_db(app_with_survey_db, mocker)
    nren_services(app_with_survey_db)
    EXCEL_FILE_ORGANISATION = openpyxl.load_workbook(os.path.join(os.path.dirname(
        __file__), "data", "2021_Organisation_DataSeries.xlsx"), data_only=True, read_only=True)
    EXCEL_FILE_USERS = openpyxl.load_workbook(os.path.join(os.path.dirname(
        __file__), "data", "2022_Connected_Users_DataSeries.xlsx"), data_only=True, read_only=True)
    EXCEL_FILE_NETWORKS = openpyxl.load_workbook(os.path.join(os.path.dirname(
        __file__), "data", "2022_Networks_DataSeries.xlsx"), data_only=True, read_only=True)
    EXCEL_FILE_NREN_SERVICES = openpyxl.load_workbook(os.path.join(os.path.dirname(
        __file__), "data", "NREN-Services-prefills_2023_Recovered.xlsx"), data_only=True, read_only=True)
    mocker.patch('compendium_v2.publishers.excel_parser.EXCEL_FILE_ORGANISATION', EXCEL_FILE_ORGANISATION)
    mocker.patch('compendium_v2.publishers.excel_parser.EXCEL_FILE_USERS', EXCEL_FILE_USERS)
    mocker.patch('compendium_v2.publishers.excel_parser.EXCEL_FILE_NETWORKS', EXCEL_FILE_NETWORKS)
    mocker.patch('compendium_v2.publishers.excel_parser.EXCEL_FILE_NREN_SERVICES', EXCEL_FILE_NREN_SERVICES)

    survey_2024 = _test_data('survey_model_2024_testdata.json')

    nren_names = ['Nren1', 'Nren2', 'Nren3', 'Nren4', 'SURF', 'KIFU', 'University of Malta', 'ASNET-AM',
                  'SIKT', 'LAT', 'RASH', 'AzScienceNet', 'GRNET', 'CSC', 'PSNC']
    with app_with_survey_db.app_context():
        db.session.add_all([presentation_models.NREN(name=nren_name, country='country') for nren_name in nren_names])
        db.session.commit()

    with app_with_survey_db.app_context():
        survey = Survey(year=2024, survey=survey_2024, status=SurveyStatus.published)
        db.session.add(survey)
        db.session.commit()

    _cli(app_with_survey_db)

    with app_with_survey_db.app_context():
        responses = db.session.query(SurveyResponse).all()
        assert len(responses) == 77

        nrens = set([resp.nren.name for resp in responses])

        assert set(nren_names).difference(nrens) == {'Nren4', 'University of Malta'}

        assert len(nrens) == len(nren_names) - 2

        for response in responses:
            data = response.answers['data']
            assert data
            for key in data:
                assert key in VALID_KEYS_2024_SURVEY
